#ifndef CvsCommand_h
#define CvsCommand_h

// $Id: CvsCommand.h,v 1.22 2013/04/29 18:27:10 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "Debug.h"

#include <QString>
#include <QStringList>
#include <QHash>

class CvsCommand: public Counter
{

    public:

    //! enumeration for command type
    enum Type
    {

        //! retrieve log for a given file (cvs log)
        LOG = 1<<0,

        //! retrieve status for a given file/directory (cvs status -l)
        STATUS = 1<<1,

        //! fake update (cvs -n update -l)
        FAKE_UPDATE = 1<<2,

        //! checkout
        CHECKOUT = 1<<3,

        //! update (cvs update -l)
        UPDATE = 1<<4,

        //! add (cvs add)
        ADD = 1<<5,

        //! remove (cvs remove)
        REMOVE = 1<<6,

        //! commit (cvs commit)
        COMMIT = 1<<7,

        //! check-out a given revision in a separate filename (cvs up -r revision -p)
        CHECKOUT_REVISION = 1<<8,

        //! any command (used to abort all)
        ANY = (CHECKOUT_REVISION << 1) - 1

    };

    //! constructor
    CvsCommand( const Type type = ANY ):
        Counter( "CvsCommand" ),
        id_( _idCounter()++ ),
        type_( type )
    {}

    //! id
    const unsigned int& id( void ) const
    { return id_; }

    //! equal to operator
    bool operator == ( const CvsCommand& command ) const
    { return id() == command.id(); }

    //! less than operator
    bool operator < ( const CvsCommand& command ) const
    { return id() < command.id(); }

    //! type
    const Type& type( void ) const
    { return type_; }

    //! initialize command names
    static void initializeNames( void );

    //! command name
    QString name( void ) const
    { return name( type() ); }

    //! map Type to command name
    static QString name( const Type& type );

    //!@name interface to arguments
    //@{

    //! add argument
    CvsCommand& addArgument( const QString& argument )
    {
        arguments_ << argument;
        return *this;
    }

    //! push arguments
    CvsCommand& operator << (const QString& argument )
    { return addArgument( argument ); }

    //! argument size
    int argumentCount( void ) const
    { return arguments_.size(); }

    //! get argument
    const QString& argument( int index ) const
    { return arguments_[index]; }

    //! clear arguments
    void clearArguments( void )
    { arguments_.clear(); }

    //@}

    //! used to retrieve commands of matching type
    class SameTypeFTor
    {
        public:

        //! constructor
        SameTypeFTor( const int& type ):
            type_( type )
        {}


        //! predicate
        bool operator() (const CvsCommand& command ) const
        { return command.type() & type_; }

        private:

        //! prediction
        int type_;

    };

    private:

    //! command names
    typedef QHash< Type, QString > NameMap;

    //! map names into type
    /*!
    note: the names are just for logging. The
    actual command is derived from the command Type
    in CvsInterface
    */
    static NameMap& _names( void );

    //! static id counter
    static unsigned int& _idCounter( void );;

    //! unique id
    unsigned int id_;

    //! type
    Type type_;

    //! arguments
    QStringList arguments_;

};

#endif
