// $Id: CvsInterface.cpp,v 1.110 2013/11/15 14:50:39 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CvsInterface.h"
#include "CvsInterface.moc"

#include "Command.h"
#include "CvsLockDialog.h"
#include "Debug.h"
#include "InformationDialog.h"
#include "IOString.h"
#include "Singleton.h"
#include "TextFormat.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QRegExp>
#include <QTextStream>

#include <algorithm>

//___________________________________________________________________
CvsInterface::CvsInterface( QObject *parent ):
    QObject( parent ),
    Counter( "CvsInterface" ),
    process_( 0 ),
    ignoreLocks_( false )
{ Debug::Throw( "CvsInterface::CvsInterface.\n" ); }

//___________________________________________________________________
void CvsInterface::logRevisions( const File& file )
{

    Debug::Throw( "CvsInterface::logRevisions.\n" );
    Q_ASSERT( !file.isDirectory() );

    _addCommand( CvsCommand( CvsCommand::LOG ) << file );
    return;

}

//_______________________________________________________________________
void CvsInterface::logStatus( const File& path, const FileInfo::List& files )
{

    Debug::Throw() << "CvsInterface::logStatus - path: " << path << endl;

    CvsCommand command = CvsCommand( CvsCommand::STATUS ) << path;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }

    _addCommand( command );

    return;

}

//_______________________________________________________________________
void CvsInterface::fakeUpdate( const File& path, const FileInfo::List& files )
{

    CvsCommand command = CvsCommand( CvsCommand::FAKE_UPDATE ) << path;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }
    _addCommand( command );

    return;

}

//_______________________________________________________________________
void CvsInterface::checkOutRevision( const File& file, const QString& revision )
{

    Debug::Throw() <<  "CvsInterface::checkOutRevision - file: " << file << " revision: " << revision << endl;
    Q_ASSERT( !file.isDirectory() );

    _addCommand( CvsCommand( CvsCommand::CHECKOUT_REVISION ) << file << revision );
    return;
}

//_______________________________________________________________________
void CvsInterface::checkOut( const File& path, const QString& cvsRoot, const QStringList& modules, const File& localPath )
{
    Debug::Throw( "CvsInterface::checkOut.\n" );

    if( modules.empty() ) return;
    else if( modules.size() == 1 )
    {
        _addCommand( CvsCommand( CvsCommand::CHECKOUT )
            << path
            << cvsRoot
            << modules.front()
            << localPath );
    } else for( QStringList::const_iterator iter = modules.begin(); iter != modules.end(); ++iter ) {

        // if several modules are provided, the localPath is ignored and modules are checked-out as-is
        _addCommand(
            CvsCommand( CvsCommand::CHECKOUT )
            << path << cvsRoot << *iter << "" );

    }

    return;
}

//_______________________________________________________________________
void CvsInterface::update( const File& path, const FileInfo::List& files )
{

    Debug::Throw( "CvsInterface::update.\n" );
    CvsCommand command = CvsCommand( CvsCommand::UPDATE ) << path;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }
    _addCommand( command );
    return;
}

//_______________________________________________________________________
void CvsInterface::add( const File& path, const FileInfo::List& files )
{

    Debug::Throw( "CvsInterface::add.\n" );
    CvsCommand command = CvsCommand( CvsCommand::ADD ) << path;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }
    _addCommand( command );
    return;
}

//_______________________________________________________________________
void CvsInterface::remove( const File& path, const FileInfo::List& files )
{

    Debug::Throw( "CvsInterface::remove.\n" );
    CvsCommand command = CvsCommand( CvsCommand::REMOVE ) << path;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }
    _addCommand( command );
    return;
}

//_______________________________________________________________________
void CvsInterface::commit( const File& path, const File& comments, const FileInfo::List& files )
{

    Debug::Throw( "CvsInterface::commit.\n" );
    CvsCommand command = CvsCommand( CvsCommand::COMMIT ) << path << comments;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }
    _addCommand( command );

    return;
}

//____________________________________________________________
bool CvsInterface::isCvsPath( File path ) const
{
    Debug::Throw( "CvsInterface::isCvsPath" );
    if( !( path.exists() && path.isDirectory() ) ) return false;
    return
        File( "CVS/Root" ).addPath( path ).exists() &&
        File( "CVS/Repository" ).addPath( path ).exists();
}

//____________________________________________________________
bool CvsInterface::isCvsFile( File file ) const
{
    Debug::Throw( "CvsInterface::isCvsFile" );
    return std::find( entries_.begin(), entries_.end(), file ) != entries_.end();
}

//____________________________________________________________
bool CvsInterface::updateRepositoryPath( File path )
{
    Debug::Throw( "CvsInterface::updateRepositoryPath.\n" );

    // try load root directory
    {
        QFile in( File( "CVS/Root" ).addPath( path ) );
        if( in.open( QIODevice::ReadOnly ) )
        {
            QString buffer( in.readAll() );
            QTextStream in_stream( &buffer );
            in_stream >> root_;
            if( in_stream.status() != QTextStream::Ok ) return false;
        } else return false;

        const QString buffer = QString( tr( "Root directory: %1\n" ) ).arg( root_ );
        emit messageAvailable( buffer, FORMAT::Bold );
    }

    // try load repository
    {
        QFile in( File( "CVS/Repository" ).addPath( path ) );
        if( in.open( QIODevice::ReadOnly ) )
        {
            QString buffer( in.readAll() );
            QTextStream in_stream( &buffer );
            in_stream >> repository_;
            if( in_stream.status() != QTextStream::Ok ) return false;
        } else return false;

        const QString buffer = QString( tr( "Repository: %1\n" ) ).arg( repository_ );
        emit messageAvailable( buffer, FORMAT::Bold );
    }

    return true;

}

//____________________________________________________________
bool CvsInterface::updateEntries( File path )
{
    Debug::Throw( "CvsInterface::updateEntries.\n" );

    // try load entries
    {

        entries_.clear();
        QFile in( File( "CVS/Entries" ).addPath( path ) );
        if( !in.open( QIODevice::ReadOnly ) ) return false;
        QStringList lines( QString( in.readAll() ).split( "\n" ) );
        for( QStringList::iterator iter = lines.begin(); iter != lines.end(); ++iter )
        {

            QStringList string_list( iter->split( "/" ) );
            if( string_list.size() < 2 ) continue;

            entries_ << File( string_list[1] ).addPath( path );

        }

    }

    return true;

}

//____________________________________________________________
void CvsInterface::abort( int type )
{
    Debug::Throw( "CvsInterface::abort.\n" );
    QList<CvsCommand>::iterator iter;

    if( !commands_.size() ) return;

    // loop over commands, keep only the ones that do not match the requested type
    QList<CvsCommand> newCommands;
    for( QList<CvsCommand>::iterator iter = commands_.begin(); iter != commands_.end(); ++iter )
    {
        if( !(iter->type() & type ) ) newCommands << *iter;
        else if( iter != commands_.begin() || _process().state() == QProcess::NotRunning ) emit commandRemoved( *iter );
    }

    // if process is running and the last command should be removed,
    // stop the process, but add the command to the list, since it is needed
    // after the process is stopped
    if( _process().state() != QProcess::NotRunning && ( commands_.front().type() & type ) )
    {
        newCommands.push_front( commands_.front() );
        commands_ = newCommands;
        const QString buffer = QString( tr( "%1 aborted.\n\n" ) ).arg( commands_.front().name() );
        emit messageAvailable( buffer );
        _process().kill();
    } else commands_ = newCommands;

    return;
}

//____________________________________________________________
void CvsInterface::abort( const CvsCommand& command )
{

    Debug::Throw() << "CvsInterface::abort - command id: " << command.id() << " type: " << command.name() << endl;
    QList<CvsCommand>::iterator iter( std::find( commands_.begin(), commands_.end(), command ) );
    if( iter == commands_.end() ) return;

    if( command == commands_.front() && _process().state() != QProcess::NotRunning )
    {

        const QString buffer = QString( tr( "%1 aborted.\n\n" ) ).arg( command.name() );
        emit messageAvailable( buffer );
        _process().kill();

    } else {
        emit commandRemoved( *iter );
        commands_.erase( iter );
    }

}

//____________________________________________________________
void CvsInterface::_error( QProcess::ProcessError error )
{

    Debug::Throw( "CvsInterface::_error.\n" );
    if( commands_.empty() ) return;

    QString buffer;
    QTextStream( &buffer ) << commands_.front().name() << ": " << CustomProcess::errorMessage( error );
    Debug::Throw() << buffer << endl;

    // it is needed to complete the command manually when
    // error message is "failed to start" otherwise the busy widget
    // will stay busy forever. Also, display a popup
    if( error == QProcess::FailedToStart )
    {
        InformationDialog( qApp->activeWindow(),  buffer ).exec();
        _commandCompleted( 0, QProcess::NormalExit );
    }

    return;
}

//___________________________________________________________________
void CvsInterface::_readStdout(  void )
{

    _process().setReadChannel( QProcess::StandardOutput );
    IOString buffer( _process() );
    stdoutBuffer_.append( buffer );
    emit messageAvailable( buffer, 0, QColor(), HIGH );
    _parseStdoutBuffer();

}

//___________________________________________________________________
void CvsInterface::_readStderr(  void )
{

    _process().setReadChannel( QProcess::StandardError );
    IOString buffer( _process() );
    stderrBuffer_.append( buffer );
    emit messageAvailable( buffer, FORMAT::Italic, QColor(), HIGH );
    _parseStderrBuffer();

}

//___________________________________________________________________
void CvsInterface::_parseStdoutBuffer( void )
{

    Debug::Throw( "CvsInterface::_parseStdoutBuffer.\n" );
    if( commands_.empty() ) return;

    // retrieve command and redirect
    // only STATUS and FAKE_UPDATE runs on fly.
    // other commands wait for the process to be completed.
    switch( commands_.front().type() )
    {

        case CvsCommand::LOG:
        _logCommandAvailable();
        break;

        case CvsCommand::STATUS:
        _statusCommandAvailable();
        break;

        case CvsCommand::FAKE_UPDATE:
        _fakeUpdateCommandAvailable();
        break;

        case CvsCommand::CHECKOUT:
        _checkOutCommandAvailable();
        break;

        case CvsCommand::UPDATE:
        _updateCommandAvailable();
        break;

        case CvsCommand::COMMIT:
        _commitCommandAvailable();
        break;

        default:
        break;
    }

}


//___________________________________________________________________
void CvsInterface::_parseStderrBuffer( void )
{

    Debug::Throw( "CvsInterface::_parseStdoutBuffer.\n" );
    if( commands_.empty() ) return;

    _checkLock( stderrBuffer_.text().mid( stderrBuffer_.position(), stderrBuffer_.text().length() ) );

    // retrieve command and redirect
    // only STATUS and FAKE_UPDATE runs on fly.
    // other commands wait for the process to be completed.
    switch( commands_.front().type() )
    {

        case CvsCommand::COMMIT:
        _commitErrorAvailable();
        break;

        default:
        break;
    }

}

//___________________________________________________________________
void CvsInterface::_commandCompleted(  int code, QProcess::ExitStatus status )
{

    Debug::Throw() << "CvsInterface::_commandCompleted - code: " << code << " status: " << status << endl;

    if( commands_.empty() ) return;
    if( status == QProcess::NormalExit )
    {

        // retrieve command and redirect
        switch( commands_.front().type() )
        {

            case CvsCommand::CHECKOUT_REVISION:
            _revisionCheckOutCommandCompleted();
            break;

            case CvsCommand::COMMIT:
            _commitCommandCompleted();
            break;

            default:
            break;
        }

    }

    if( status == QProcess::NormalExit )
    {
        const QString buffer = QString( tr( "%1 completed.\n\n" ) ).arg( commands_.front().name() );
        emit messageAvailable( buffer );
    }

    //! emit signal
    emit commandCompleted( commands_.front() );

    // remove command from list
    // run next command if any
    commands_.removeFirst();
    if( !commands_.empty() ) _run();

    return;

}

//____________________________________________________________
void CvsInterface::_resetBuffers( void )
{
    Debug::Throw( "CvsInterface::_resetBuffers.\n" );
    stdoutBuffer_.clear();
    stderrBuffer_.clear();
}

//_______________________________________________________________________
void CvsInterface::_addCommand( CvsCommand command )
{

    // add to list, emit signal, and run.
    commands_ << command;
    emit commandAdded( command );

    if( commands_.size() == 1 ) _run();
}

//_______________________________________________________________________
void CvsInterface::_run( void )
{
    Debug::Throw( "CvsInterface::_run.\n" );
    if( commands_.empty() ) return;

    // reset buffer
    _resetBuffers();
    _resetProcess();

    // reset ignoreLocks_ flag
    ignoreLocks_ = false;

    // retrieve pending command
    const CvsCommand& command( commands_.front() );
    emit commandStarted( command );

    // retrieve cvs command from options
    Command cvs( XmlOptions::get().raw( "CVS" ) );

    // start process with matching command type.
    switch (command.type())
    {

        case CvsCommand::LOG:
        {
            const QString buffer = QString( "%1 %2\n" ).arg( command.name() ).arg( command.argument(0) );
            emit messageAvailable( buffer, FORMAT::Bold );

            _newProcess().setWorkingDirectory( QFileInfo( command.argument(0) ).absolutePath() );
            _process().start( cvs << "log" << QFileInfo( command.argument(0) ).fileName() );

        }
        break;

        case CvsCommand::STATUS:
        { _runStatusCommand(); }
        break;

        case CvsCommand::FAKE_UPDATE:
        { _runFakeUpdateCommand(); }
        break;

        case CvsCommand::CHECKOUT_REVISION:
        {

            const QString buffer = QString( "%1 %2\n" ).arg( command.name() ).arg( command.argument(0) );
            emit messageAvailable( buffer, FORMAT::Bold );

            _newProcess().setWorkingDirectory( QFileInfo( command.argument(0) ).absolutePath() );
            _process().start(
                cvs << "update"
                << "-r" << command.argument(1)
                << "-p" << QFileInfo( command.argument(0) ).fileName() );
        }
        break;

        case CvsCommand::CHECKOUT:
        { _runCheckOutCommand(); }
        break;

        case CvsCommand::UPDATE:
        { _runUpdateCommand(); }
        break;

        case CvsCommand::ADD:
        {
            const QString buffer = QString( "%1 %2\n" ).arg( command.name() ).arg( command.argument(0) );
            emit messageAvailable( buffer, FORMAT::Bold );

            QString path( command.argument(0) );
            Q_ASSERT( QFileInfo( path ).isDir() );
            _newProcess().setWorkingDirectory( path );

            QStringList commandString( cvs << "add" );
            for( int i=1; i<command.argumentCount(); i++ )
            { commandString << QFileInfo( command.argument(i) ).fileName(); }

            _process().start( commandString );

        }
        break;

        case CvsCommand::REMOVE:
        {
            const QString buffer = QString( "%1 %2\n" ).arg( command.name() ).arg( command.argument(0) );
            emit messageAvailable( buffer, FORMAT::Bold );

            QString path( command.argument(0) );
            Q_ASSERT( QFileInfo( path ).isDir() );
            _newProcess().setWorkingDirectory( path );

            QStringList commandString( cvs << "remove" );
            for( int i=1; i<command.argumentCount(); i++ )
            { commandString << QFileInfo(command.argument(i)).fileName(); }

            _process().start( commandString );

        }
        break;

        case CvsCommand::COMMIT:
        { _runCommitCommand(); }
        break;

        default: break;

    }

    return;

}

//_______________________________________________________________________
void CvsInterface::_runStatusCommand( void )
{
    // retrieve pending command
    CvsCommand& command( commands_.front() );

    // retrieve path for update command
    QString path( command.argument(0) );
    Q_ASSERT( QFileInfo(path).isDir() );

    const QString buffer = QString( "%1 %2\n" ).arg( command.name() ).arg( command.argument(0) );
    emit messageAvailable( buffer, FORMAT::Bold );

    // do nothing if path is not CVS
    if( !isCvsPath( File( path ) ) )  _commandCompleted( 0, QProcess::NormalExit );
    else
    {
        // retrieve cvs command
        Command cvs( XmlOptions::get().raw( "CVS" ) );
        _newProcess().setWorkingDirectory( path );
        QStringList commandString( cvs << "status" << "-l" );
        for( int i=1; i<command.argumentCount(); i++ )
        {

            // no need to run on subdirectories, since they will not
            // appear in the file browser anyway
            QFileInfo file( command.argument(i) );
            if( !file.isDir() )
            { commandString << file.fileName(); }

        }

        _process().start( commandString );
    }

}

//_______________________________________________________________________
void CvsInterface::_runFakeUpdateCommand( void )
{
    // retrieve pending command
    CvsCommand& command( commands_.front() );

    // retrieve path for update command
    QString path( command.argument(0) );
    Q_ASSERT( QFileInfo( path ).isDir() );

    const QString buffer = QString( "%1 %2\n" ).arg( command.name() ).arg( command.argument(0) );
    emit messageAvailable( buffer, FORMAT::Bold );

    // do nothing if path is not CVS
    if( !isCvsPath( path ) )  _commandCompleted( 0, QProcess::NormalExit );
    else
    {
        // retrieve cvs command
        Command cvs( XmlOptions::get().raw( "CVS" ) );
        _newProcess().setWorkingDirectory( path );
        QStringList commandString( cvs << "-n" << "update" << "-l" );
        for( int i=1; i<command.argumentCount(); i++ )
        {
            // no need to run on subdirectories, since they will not
            // appear in the file browser anyway
            QFileInfo file( command.argument(i) );
            if( !file.isDir() )
            { commandString << file.fileName(); }
        }
        _process().start( commandString );

    }

    return;
}

//_______________________________________________________________________
void CvsInterface::_runCheckOutCommand( void )
{
    Debug::Throw( "CvsInterface::_runCheckOutCommand.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );
    Command cvs( XmlOptions::get().raw( "CVS" ) );

    // retrieve path for update command
    QString path( command.argument(0) );
    Q_ASSERT( QFileInfo( path ).isDir() );

    QString cvsRoot( command.argument(1) );
    QString module( command.argument(2) );
    QString localPath( command.argument(3) );

    const QString buffer = QString( "%1 %2\n" ).arg( command.name() ).arg( module );
    emit messageAvailable( buffer );

    QStringList commandString( cvs );
    if( !cvsRoot.isEmpty() ) commandString << "-d" << cvsRoot;
    commandString << "checkout";

    if( !localPath.isEmpty() ) commandString << "-d" << localPath;
    commandString << module;

    _newProcess().setWorkingDirectory( path );
    _process().start( commandString );

    return;
}


//_______________________________________________________________________
void CvsInterface::_runUpdateCommand( void )
{
    Debug::Throw( "CvsInterface::_runUpdateCommand.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );
    Command cvs( XmlOptions::get().raw( "CVS" ) );

    // retrieve path for update command
    QString path( command.argument(0) );
    Q_ASSERT( QFileInfo( path ).isDir() );

    const QString buffer = QString( "%1 %2\n" ).arg( command.name() ).arg( path );
    emit messageAvailable( buffer );

    // directories that are found in the arguments
    // must be handled separately by CVS.
    // they are kept aside and processed via dedicated commands
    // added (recursively) from here .
    QStringList subdirectories;
    QStringList files;
    for( int i=1; i<command.argumentCount(); i++ )
    {
        QString file( command.argument(i) );
        if( QFileInfo( file ).isDir() )
        {
            Debug::Throw() << "CvsInterface::_runUpdateCommand - subdirectory: " << file << endl;
            subdirectories << file;
        } else files << file;
    }

    // run current command if any files are in the arguments
    // otherwise set command as completed
    if( !files.empty() || subdirectories.empty() )
    {

        // prepare command
        QStringList commandString( cvs << "update" );

        // add files
        for( QStringList::iterator iter = files.begin(); iter != files.end(); ++iter )
        { commandString << QFileInfo( *iter ).fileName(); }

        // run process
        _newProcess().setWorkingDirectory( path );
        _process().start( commandString );

    } else _commandCompleted( 0, QProcess::NormalExit );

    // add dedicated commands for found subdirectories
    if( !subdirectories.empty() )
    {
        for( QStringList::iterator iter = subdirectories.begin(); iter != subdirectories.end(); ++iter )
        {
            // need to make sure the trailing "/" is needed or not.
            _addCommand( CvsCommand( CvsCommand::UPDATE ) << (*iter + "/") );
        }
    }

    return;

}

//______________________________________________________________
void CvsInterface::_runCommitCommand( void )
{
    Debug::Throw( "CvsInterface::_runCommitCommand.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );
    Command cvs( XmlOptions::get().raw( "CVS" ) );

    // retrieve path for update command
    QString path = command.argument(0);
    Q_ASSERT( QFileInfo( path ).isDir() );

    // try retrieve comments file
    QString comments( command.argument(1) );
    const QString buffer = QString( "%1 %2\n" ).arg( command.name() ).arg( path );
    emit messageAvailable( buffer, FORMAT::Bold );

    // directories that are found in the arguments
    // must be handled separately by CVS.
    // they are kept aside and processed via dedicated commands
    // added (recursively) from here .
    QStringList subdirectories;
    QStringList files;
    for( int i=2; i<command.argumentCount(); i++ )
    {
        QString file( command.argument(i) );
        if( QFileInfo( file ).isDir() )
        {

            Debug::Throw() << "CvsInterface::_runCommitCommand - subdirectory: " << file << endl;
            subdirectories << file;

        } else files << file;

    }

    // run current command if any files are in the arguments
    // otherwise set command as completed
    if( !files.empty() || subdirectories.empty() )
    {

        // prepare command
        QStringList commandString( cvs << "commit" );

        // add comments file, if any
        if( comments.isEmpty() || !QFileInfo( comments ).exists() ) commandString << "-m" << "\"no message\"";
        else commandString << "-F" << comments;

        // add files
        for( QStringList::iterator iter = files.begin(); iter != files.end(); ++iter )
        { commandString << QFileInfo( *iter ).fileName(); }

        // run process
        _newProcess().setWorkingDirectory( path );
        _process().start( commandString );

    } else _commandCompleted( 0, QProcess::NormalExit );

    // add dedicated commands for found subdirectories
    if( !subdirectories.empty() )
    {
        for( QStringList::iterator iter = subdirectories.begin(); iter != subdirectories.end(); ++iter )
        {
            // need to make sure the trailing slash is needed
            _addCommand( CvsCommand( CvsCommand::COMMIT ) << (*iter + "/") << comments );
        }
    }

}

//_______________________________________________________________________
bool CvsInterface::_checkLock( const QString& buffer )
{
    Debug::Throw( "CvsInterface::_checkLock.\n" );

    static const QString userStartTag = "waiting for";
    static const QString pathStartTag = "'s lock in";
    static const QString endOfLine = "\n";

    int userStart( buffer.indexOf( userStartTag ) );
    if( userStart < 0 ) return false;
    userStart += userStartTag.length();

    int pathStart( buffer.indexOf( pathStartTag, userStart ) );
    if( pathStart < 0 ) return false;

    // retrieve user
    QString user( buffer.mid( userStart, pathStart - userStart ).trimmed() );
    pathStart += pathStartTag.length();

    int pathEnd( buffer.indexOf( endOfLine, pathStart ) );
    if( pathEnd < 0 ) return false;

    // parse path
    QString path( buffer.mid( pathStart, pathEnd - pathStart ).trimmed() );

    // emit message
    {
        const QString buffer = QString( tr( "Found lock from '%1' in '%2'\n" ) ).arg( user ).arg( path );
        emit messageAvailable( buffer );
    }

    // open dialog
    if( !ignoreLocks_ )
    {
        // set ignore_locks to true (to avoid multiple dialogs to be oppened)
        // opens CvsLockDialog.
        // If dialog is accepted, abort all actions.
        ignoreLocks_ = true;
        if( CvsLockDialog( qApp->activeWindow(), user, path ).exec() == QDialog::Accepted ) abort();
    }

    return true;

}

//___________________________________________________________________
void CvsInterface::_logCommandAvailable( void )
{

    // need to check local vs absolute names

    Debug::Throw( "CvsInterface::_logCommandAvailable.\n" );

    // parsing strings
    static const QRegExp separator( "\\s(-{28,})|(={28,})\\s" );
    static const QString revisionTag( "revision" );
    static const QString dateTag( "date:" );
    static const QString authorTag( "author:" );
    static const QString tagEnd( ";" );

    // look for first separator in full buffer. It is needed because at least one separator is needed
    // to start the revisions
    int firstSeparator( stdoutBuffer_.text().indexOf( separator ) );
    if( firstSeparator < 0 ) return;

    int start = qMax( stdoutBuffer_.position(), firstSeparator + separator.matchedLength() );

    Revision revision;
    Revision::List revisions;
    while(1)
    {

        // allocate current revision
        Revision revision;

        // look for revision
        int revisionStart = stdoutBuffer_.text().indexOf( revisionTag, start );
        if( revisionStart < 0 ) break;
        revisionStart += revisionTag.length();

        // look for end of line
        int endOfLine = stdoutBuffer_.text().indexOf( "\n", revisionStart );
        if( endOfLine < 0 ) break;

        // parse revision name
        QString revision_name = stdoutBuffer_.text().mid( revisionStart, endOfLine - revisionStart ).trimmed();
        revision.setName( revision_name );

        // look for date
        int dateStart = stdoutBuffer_.text().indexOf( dateTag, endOfLine );
        if( dateStart < 0 ) break;
        dateStart += dateTag.length();

        int dateEnd = stdoutBuffer_.text().indexOf( tagEnd, dateStart );
        if( dateEnd < 0 ) break;

        // parse date
        QString date_string = stdoutBuffer_.text().mid( dateStart, dateEnd-dateStart ).trimmed();
        revision.setTimeStamp( date_string );

        // look for author
        int authorStart = stdoutBuffer_.text().indexOf( authorTag, dateEnd );
        if( authorStart < 0 ) break;
        authorStart += authorTag.length();

        int authorEnd = stdoutBuffer_.text().indexOf( tagEnd, authorStart );
        if( authorEnd < 0 ) break;

        // parse author
        QString author = stdoutBuffer_.text().mid( authorStart, authorEnd - authorStart ).trimmed();
        revision.setAuthor( author );

        // look for next end of line to start reading comments
        int commentsStart = stdoutBuffer_.text().indexOf( "\n", authorEnd );
        if( commentsStart < 0 ) break;
        commentsStart++;

        // look for next separator for end of comments
        int commentsEnd = stdoutBuffer_.text().indexOf( separator, commentsStart );
        if( commentsEnd < 0 ) break;

        // parse comments
        QString comments = stdoutBuffer_.text().mid( commentsStart, commentsEnd - commentsStart );
        revision.setComments( comments );

        // store revision and emit message
        revisions << revision;
        QString message = QString( tr( "revision: %1 date: %2 author: %3\n" ) )
            .arg( revision.name() )
            .arg( revision.timeStamp().toString() )
            .arg( revision.author() );

        emit messageAvailable( message );

        // update flags
        start = commentsEnd + separator.matchedLength();

    }

    // update position in buffer
    stdoutBuffer_.flush( start );

    // send found revisions and return
    if( !revisions.empty() )
    { emit revisionsAvailable( commands_.front().argument(0), revisions ); }

    return;

}

//___________________________________________________________________
void CvsInterface::_statusCommandAvailable( void )
{

    Debug::Throw( "CvsInterface::_statusCommandAvailable.\n" );

    // map cvs tags versus fileInfo status
    static QHash< QString, FileInfo::Status > statusMap;
    if( statusMap.empty() )
    {
        statusMap["Up-to-date"] = FileInfo::UpToDate;
        statusMap["Locally Modified"] = FileInfo::LocallyModified;
        statusMap["Locally Added"] = FileInfo::LocallyAdded;
        statusMap["Locally Removed"] = FileInfo::LocallyRemoved;
        statusMap["Needs Patch"] = FileInfo::RemotelyModified;
        statusMap["Entry Invalid"] = FileInfo::RemotelyRemoved;
        statusMap["Needs Merge"] = FileInfo::Merge;
        statusMap["Needs Checkout"] = FileInfo::RemotelyAdded;
        statusMap["File had Conflicts on Merge"] = FileInfo::Conflict;
    }

    QRegExp fileTag( "File: (no file )?" );
    static const QString statusTag = "Status: ";
    static const QString endOfLine = "\n";

    // list of loaded information
    QString path( commands_.front().argument(0) );
    FileInfo::List files;

    // parse buffer
    while(1)
    {

        // look for file tag
        int file_pos = stdoutBuffer_.text().indexOf( fileTag, stdoutBuffer_.position() );
        if( file_pos < 0 ) break;

        int position( file_pos + fileTag.matchedLength() );

        // look for status tag
        int status_pos = stdoutBuffer_.text().indexOf( statusTag, position );
        if( status_pos < 0 ) break;

        // look for next end of line
        int end_pos = stdoutBuffer_.text().indexOf( endOfLine, status_pos );
        if( end_pos < 0 ) break;

        // retrieve file and status
        QString file = QString( stdoutBuffer_.text() ).mid( position, status_pos - position ).trimmed();
        QString status = QString( stdoutBuffer_.text() ).mid( status_pos + statusTag.size(), end_pos - status_pos - statusTag.size() ).trimmed();

        QHash< QString, FileInfo::Status >::iterator status_iter( statusMap.find( status ) );
        files << FileInfo( File(file).addPath( path ), FileInfo::Document, status_iter.value() );

        QString buffer;
        QTextStream(&buffer) << status << " - " << file << endl;
        emit messageAvailable( buffer );

        // update position for next bunch of available data
        stdoutBuffer_.flush( end_pos + endOfLine.size() );

    }

    emit statusAvailable( path, files );
    return;

}

//___________________________________________________________________
void CvsInterface::_fakeUpdateCommandAvailable( void )
{
    Debug::Throw( "CvsInterface::_fakeUpdateCommandAvailable.\n" );

    static const QRegExp tag( "\\b(U|P|C|M)\\b" );
    static const QString endOfLine( "\n" );

    static QHash<QString, QString> tagMap;
    static QHash<QString, FileInfo::Status> statusMap;

    if( tagMap.empty() )
    {
        tagMap["U"] = tr( "Remotely Added" );
        tagMap["P"] = tr( "Remotely Modified" );
        tagMap["M"] = tr( "Locally Modified" );
        tagMap["C"] = tr( "Conflict" );
    }

    if( statusMap.empty() )
    {
        statusMap["U"] = FileInfo::RemotelyModified;
        statusMap["P"] = FileInfo::RemotelyModified;
        statusMap["M"] = FileInfo::LocallyModified;
        statusMap["C"] = FileInfo::Conflict;
    }

    // list of loaded information
    QString path( commands_.front().argument(0) );
    FileInfo::List files;
    while(1)
    {

        int startPos( tag.indexIn( stdoutBuffer_.text().left( stdoutBuffer_.position() ) ) );
        if( startPos < 0 ) break;

        int length( tag.matchedLength() );

        // parse tag and lookup in conversion table
        QString tagMatch = stdoutBuffer_.text().mid( startPos, length );

        QHash<QString, FileInfo::Status>::iterator status_iter( statusMap.find( tagMatch ) );
        Q_ASSERT( status_iter != statusMap.end() );
        FileInfo::Status status( status_iter.value() );

        // convert tag to readable string
        QHash<QString, QString>::iterator tagIter( tagMap.find( tagMatch ) );
        Q_ASSERT( tagIter != tagMap.end() );
        tagMatch = tagIter.value();

        // look for file
        stdoutBuffer_.flush( startPos + length );
        int end( stdoutBuffer_.text().indexOf( endOfLine, stdoutBuffer_.position() ) );
        if( end < 0 ) break;

        QString file = stdoutBuffer_.text().mid( stdoutBuffer_.position(), end-stdoutBuffer_.position() ).trimmed();
        files << FileInfo( File(file).addPath( path ), FileInfo::Document, status );

        const QString buffer = QString( "%1 - %2\n" ).arg( tagMatch ).arg( file );
        emit messageAvailable( buffer );

        stdoutBuffer_.flush( end );

    }

    emit statusAvailable( commands_.front().argument(0), files );
    return;
}

//___________________________________________________________________
void CvsInterface::_revisionCheckOutCommandCompleted( void )
{
    Debug::Throw( "CvsInterface::_revisionCheckOutCommandCompleted.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );

    // generate destination filename from source and revision name
    File destination( command.argument(0) );
    QString rev = QString( command.argument(1) ).replace( " ", "_" );

    destination = destination.truncatedName() + "-" + rev + "." + destination.extension();

    // write buffer to destination file
    QFile out( destination );
    if( !out.open( QIODevice::WriteOnly ) )
    {
        const QString buffer = QString( tr( "Unable to write to '%1'. <Cvs Up> canceled." ) ).arg( destination );
        InformationDialog( qApp->activeWindow(), buffer ).exec();
        return;
    }

    // retrieve all stdout and put into stream
    out.write( stdoutBuffer_.text().toLatin1() );
    out.close();

    const QString buffer = QString( tr( "%1 created.\n" ) ).arg( destination );
    emit messageAvailable( buffer );
    emit revisionAvailable( command.argument(0), destination );
    emit scratchFileAdded( File( destination ) );

}

//___________________________________________________________________
void CvsInterface::_checkOutCommandAvailable( void )
{
    Debug::Throw( "CvsInterface::_checkOutCommandAvailable.\n" );

    static const QRegExp tag( "\\b(U)\\b" );
    static const QString endOfLine( "\n" );
    static QHash<QString, QString> tagMap;

    if( tagMap.empty() )
    {
        tagMap["U"] = tr( "Added" );
    }

    // list of loaded information
    FileInfo::List files;
    while(1)
    {

        // look for tag
        int startPos( tag.indexIn( stdoutBuffer_.text().left( stdoutBuffer_.position() ) ) );
        if( startPos < 0 ) break;
        int length( tag.matchedLength() );
        if( length > 1 )
        {
            startPos ++;
            length --;
        }

        // parse tag and lookup in conversion table
        QString tag = stdoutBuffer_.text().mid( startPos, length );
        QHash<QString, QString>::iterator iter( tagMap.find( tag ) );
        Q_ASSERT( iter != tagMap.end() );
        tag = iter.value();

        // look for file
        stdoutBuffer_.flush( startPos + length );
        int end( stdoutBuffer_.text().indexOf( endOfLine, stdoutBuffer_.position() ) );
        if( end < 0 ) break;

        QString file = stdoutBuffer_.text().mid( stdoutBuffer_.position(), end-stdoutBuffer_.position() ).trimmed();
        files << FileInfo( file, FileInfo::Document, FileInfo::UpToDate );

        const QString buffer = QString( "%1 - %2\n" ).arg( tag ).arg( file );
        emit messageAvailable( buffer );

        stdoutBuffer_.flush( end + endOfLine.size() );

    }

    emit statusAvailable( commands_.front().argument(0), files );
    return;

};

//___________________________________________________________________
void CvsInterface::_updateCommandAvailable( void )
{
    Debug::Throw( "CvsInterface::_updateCommandAvailable.\n" );
    static const QRegExp tag( "\\b(U|P|M|C|A|R)\\b" );
    static const QString endOfLine( "\n" );
    static QHash<QString, QString> tagMap;
    static QHash<QString, FileInfo::Status> statusMap;

    if( tagMap.empty() )
    {
        tagMap["U"] = tr( "Updated" );
        tagMap["P"] = tr( "Updated" );
        tagMap["M"] = tr( "Locally Modified" );
        tagMap["C"] = tr( "Conflict" );
        tagMap["A"] = tr( "Locally Added" );
        tagMap["R"] = tr( "Locally Removed" );
    }

    // maps the tag to the status of the file *after* the update has been performed
    if( statusMap.empty() )
    {
        statusMap["U"] = FileInfo::UpToDate;
        statusMap["P"] = FileInfo::UpToDate;
        statusMap["M"] = FileInfo::LocallyModified;
        statusMap["C"] = FileInfo::Conflict;
        statusMap["A"] = FileInfo::LocallyAdded;
        statusMap["R"] = FileInfo::LocallyRemoved ;
    }

    FileInfo::List files;
    while(1)
    {

        // look for tag
        int startPos( tag.indexIn( stdoutBuffer_.text().left( stdoutBuffer_.position() ) ) );
        if( startPos < 0 ) break;

        int length( tag.matchedLength() );

        // parse tag and lookup in conversion table
        QString tag = stdoutBuffer_.text().mid( startPos, length );
        QHash<QString, FileInfo::Status>::iterator status_iter( statusMap.find( tag ) );
        FileInfo::Status status( status_iter.value() );

        // convert tag to readable string
        QHash<QString, QString>::iterator tagIter( tagMap.find( tag ) );
        Q_ASSERT( tagIter != tagMap.end() );
        tag = tagIter.value();

        // look for file
        stdoutBuffer_.flush( startPos + length );
        int end( stdoutBuffer_.text().indexOf( endOfLine, stdoutBuffer_.position() ) );
        if( end < 0 ) break;

        QString file = stdoutBuffer_.text().mid( stdoutBuffer_.position(), end-stdoutBuffer_.position() ).trimmed();
        files << FileInfo( file, FileInfo::Document, status );

        const QString buffer = QString( "%1 - %2\n" ).arg( tag ).arg( file );
        emit messageAvailable( buffer );

        stdoutBuffer_.flush( end + endOfLine.size() );

    }

    emit statusAvailable( commands_.front().argument(0), files );
    return;

}

//___________________________________________________________________
void CvsInterface::_commitCommandAvailable( void )
{
    Debug::Throw( "CvsInterface::_commitCommandAvailable.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );

    // look for committed files
    static const QRegExp commitTag( "(Checking in)|(Removing)" );
    static const QString endOfLine( ";\n" );
    FileInfo::List files;
    while(1)
    {

        // look for tag
        int startPos( commitTag.indexIn( stdoutBuffer_.text().left( stdoutBuffer_.position() ) ) );
        if( startPos < 0 ) break;

        // look for file
        stdoutBuffer_.flush( startPos + commitTag.matchedLength() );
        int end( stdoutBuffer_.text().indexOf( endOfLine, stdoutBuffer_.position() ) );
        if( end < 0 ) break;

        QString file = stdoutBuffer_.text().mid( stdoutBuffer_.position(), end-stdoutBuffer_.position() ).trimmed();
        files << FileInfo( file, FileInfo::Document, FileInfo::UpToDate );

        const QString buffer = QString( tr( "Committed '%1'\n" ) ).arg( File( file ).addPath( command.argument(0) ) );
        emit messageAvailable( buffer );

        stdoutBuffer_.flush( end + endOfLine.size() );

    }

    emit statusAvailable( commands_.front().argument(0), files );
    return;

}

//___________________________________________________________________
void CvsInterface::_commitErrorAvailable( void )
{
    Debug::Throw( "CvsInterface::_commitErrorAvailable.\n" );

    // here you could parse the output and log important informations
    // look for commit aborted
    static const QString errorTag( "cvs [commit aborted]" );
    static const QString previousTag( "cvs commit:" );

    // look for error message
    int errorPos( stderrBuffer_.text().indexOf( errorTag, stderrBuffer_.position() ) );
    if( errorPos < 0 ) return;

    // look for previous tag
    QString buffer;
    int previousPos( stderrBuffer_.text().lastIndexOf( previousTag, errorPos ) );
    if( previousPos < 0 )
    {

        buffer = QString(
            tr( "Commit aborted for an unknown reason.\n"
            "Retry with log frame visible and maximum verbosity\n"
            "to get more information." ) );

    } else {

        buffer = QString(
            tr( "Commit aborted for the following reason:\n %1\n"
            "Fix the error and retry.\n" ) ).arg( stderrBuffer_.text().mid( previousPos + previousTag.length(), errorPos - (previousPos + previousTag.length()) ).trimmed() );

    }

    InformationDialog(  qApp->activeWindow(), buffer ).exec();
    emit messageAvailable( buffer, FORMAT::Italic );
    stderrBuffer_.flush( errorPos + errorTag.length() );

    return;

}

//___________________________________________________________________
void CvsInterface::_commitCommandCompleted( void )
{
    Debug::Throw( "CvsInterface::_commitCommandCompleted.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );

    // retrieve temporary file and try remove
    File tmp( command.argument(1) );
    if( tmp.exists() ) tmp.remove();

}

//___________________________________________
void CvsInterface::_resetProcess( void )
{
    if( process_ )
    {
        process_->deleteLater();
        process_ = 0;
    }
}

//____________________________________________________________
CustomProcess& CvsInterface::_newProcess( void )
{

    // delete old process
    _resetProcess();

    // create new one
    process_ = new CustomProcess( this );
    connect( &_process(), SIGNAL(finished(int,QProcess::ExitStatus)), this, SLOT(_commandCompleted(int,QProcess::ExitStatus)) );
    connect( &_process(), SIGNAL(error(QProcess::ProcessError)), SLOT(_error(QProcess::ProcessError)) );
    connect( &_process(), SIGNAL(readyReadStandardOutput()), SLOT(_readStdout()) );
    connect( &_process(), SIGNAL(readyReadStandardError()), SLOT(_readStderr()) );

    return _process();

}
