// $Id: FileInfo.cpp,v 1.23 2013/04/29 18:27:10 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "FileInfo.h"
#include "Debug.h"
#include "TimeStamp.h"

#include <QStringList>

//! Some Xml definitions
namespace XML
{
    static const QString CREATION( "Creation" );
    static const QString STATUS( "Status" );
}

//________________________________________________________________
FileInfo::FileInfo( File file, TypeFlags type ):
    BaseFileInfo( file, type ),
    Counter( "FileInfo" ),
    dirty_( true ),
    status_( Unknown ),
    creation_( TimeStamp::now() )
{}

//________________________________________________________________
FileInfo::FileInfo( File file, TypeFlags type, StatusFlags status ):
    BaseFileInfo( file, type ),
    Counter( "FileInfo" ),
    dirty_( true ),
    status_( status ),
    creation_( TimeStamp::now() )
{}

//________________________________________________________________
FileInfo::FileInfo( const QDomElement& element ):
    BaseFileInfo( element ),
    Counter( "FileInfo" ),
    dirty_( true ),
    status_( Unknown ),
    creation_(0)
{

    Debug::Throw( "FileInfo::FileInfo (dom).\n" );

    // parse attributes
    QDomNamedNodeMap attributes( element.attributes() );
    for( unsigned int i=0; i<attributes.length(); i++ )
    {
        QDomAttr attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        else if( attribute.name() == XML::STATUS ) setStatus( StatusFlags( attribute.value().toUInt() ) );
        else if( attribute.name() == XML::CREATION ) setCreation( attribute.value().toInt() );
    }

}

//________________________________________________________________
QDomElement FileInfo::domElement( QDomDocument& document ) const
{
    Debug::Throw( "FileInfo::DomElement.\n" );
    QDomElement out( BaseFileInfo::domElement( document ) );
    out.setAttribute( XML::STATUS, QString().setNum( status() ) );
    out.setAttribute( XML::CREATION, QString().setNum( created() ) );
    return out;
}

//________________________________________________________________
QString FileInfo::statusString( StatusFlags flags )
{

    QStringList statusList;
    if( flags & UpToDate ) statusList << QObject::tr( "Up to date" );
    if( flags & LocallyModified ) statusList << QObject::tr( "Needs commit (modified locally)" );
    if( flags & LocallyAdded ) statusList << QObject::tr( "Needs commit (new file)" );
    if( flags & LocallyRemoved ) statusList << QObject::tr( "Needs commit (removed locally)" );
    if( flags & RemotelyModified ) statusList << QObject::tr( "Needs update (modified remotely)" );
    if( flags & RemotelyAdded ) statusList << QObject::tr( "Needs update (new file in repository)" );
    if( flags & RemotelyRemoved ) statusList << QObject::tr( "Needs update (removed from repository)" );
    if( flags & Merge ) statusList << QObject::tr( "Needs merge with repository version" );
    if( flags & Conflict ) statusList << QObject::tr( "Conflicts with repository version" );
    if( statusList.isEmpty() ) return QString();
    else return statusList.join( " " );

}

//________________________________________________________________
void FileInfo::update( void )
{

    if( dirty() )
    {
        BaseFileInfo::update();
        setDirty( false );
    }

}

//________________________________________________________________
void FileInfo::updateFrom( const FileInfo& info )
{

    if( dirty() )
    {
        Q_ASSERT( file() == info.file() );
        Q_ASSERT( !info.dirty() );
        BaseFileInfo::updateFrom( info );
        setDirty( false );
    }

}
